#!/usr/bin/python3
import os
import sys
import time
from gi.repository import GLib
from pydbus import SystemBus
from pydbus.generic import signal
import importlib,sys
importlib.reload(sys)
import dbus
import dbus.service
from dbus.mainloop.glib import DBusGMainLoop
import threading
import socket
import time
import random
from datetime import datetime
from threading import Timer
import subprocess
import json


def get_host_name():
    """获取当前用户名
    :return: username
    """
    try:
        bus2 = dbus.SystemBus()
        proxy = bus2.get_object("com.kylin.kydevmonit.hedronclient","/kydevmonit/hedronclient")
        username = proxy.get_current_user()
        if(username == '' or username == ' '):
            username = '未登录'
    except:
        username = '未登录'
    return  username

def get_host_ip():
    """
    查询本机ip地址
    :return: ip
    """
    try:
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        s.connect(('8.8.8.8', 80))
        ip = s.getsockname()[0]
    except Exception:
        return "127.0.0.1"
    return ip

def get_host_time():
    """获取本机时间
    :return: localtime
    """
    localtime = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
    return localtime

def writelog(message):
    """写本地日志"""
    try:
        fo = open("/var/log/netIsolation.log", "a")
        localtime = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
        fo.write(localtime + " [info] "+ message +"\n")
        fo.close()
    except:
        print("写日志失败")

def passlog(message):
    """写审计日志"""
    info = {
        "time":get_host_time(),
        "hostname":get_host_name(),
        "ip":get_host_ip(),
        "name":"Net Isolation Control",
        "lv":"error",
        "message":message
    }
    pass_info = json.dumps(info, ensure_ascii=False)
    writelog(pass_info)
    try:
        log_object = SystemBus().get("org.log.sys_transmit","/org/log/sys_transmit")
        ret = log_object.log_transmit(pass_info)
        writelog("sent log ret = {}".format(ret))
    except:
        writelog("发送审计日志失败")

def get_status():
    """获取管控状态"""
    try:
        with open("/etc/net_isolation.conf","r",encoding="utf-8") as file:
            status = 1
    except:
        status = 0
        writelog("/etc/net_isolation.conf 文件不存在")
    return status

def get_net_time():
    """获取定时检测时间"""
    try:
        with open("/etc/net_isolation_time.conf","r",encoding="utf-8") as file:
            net_time = file.read()
    except:
        net_time = 0
        writelog("/etc/net_isolation_time.conf 文件不存在")
    return net_time

def get_in_ip():
    """获取配置的内网ip"""
    try:
        with open("/etc/net_isolation_ip.conf","r",encoding="utf-8") as file:
            in_ip = file.read().replace("\n",",")
            in_ip_list = in_ip.split(",")
    except:
        in_ip_list=[]
        writelog("/etc/net_isolation_ip 文件不存在")
    return in_ip_list

def get_out_ip():
    """获取配置的外网ip"""
    try:
        with open("/etc/net_isolation_ip_out.conf","r",encoding="utf-8") as file:
            in_ip = file.read().replace("\n",",")
            in_ip_list = in_ip.split(",")
    except:
        in_ip_list=[]
        writelog("/etc/net_isolation_ip_out 文件不存在")
    return in_ip_list

def get_dns():
    """获取配置的内网dns"""
    try:
        with open("/etc/net_isolation_dns.conf","r",encoding="utf-8") as file:
            dns = file.read().replace("\n",",")
            dns_list = dns.split(",")
    except:
        dns_list=[]
        writelog("/etc/net_isolation_dns.conf 文件不存在")
    return dns_list

def get_devices():
    """获取所有网卡设备"""
    netcard_list = [] #存放所有网卡名
    net_object = SystemBus().get("org.freedesktop.NetworkManager", "/org/freedesktop/NetworkManager")
    devices = net_object.AllDevices
    for device in devices:
        device_object = SystemBus().get("org.freedesktop.NetworkManager", device)
        netcard = device_object.Interface
        netcard_list.append(netcard)
    return netcard_list

def check_in_net(netcard):
    """检查内网连通性"""
    ret = 0
    in_ip_list = get_in_ip()

    random.shuffle(in_ip_list)

    for ip in in_ip_list:
        writelog("ping " + ip + " -c 1 -w 5 -I " + netcard)
        command = ["/usr/bin/ping", ip, "-c", "1", "-w", "5", "-I", netcard]
        try:
            result = subprocess.run(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
            if result.returncode == 0: #能ping通
                ret = 1
                break;
        except Exception as e:
            print("Error executing command:", e)

    writelog("网卡 {} 能否联通内网: {}".format(netcard, ret))
    return ret

def check_out_net(netcard):
    """检查外网连通性"""
    ret = 0
    out_ip_list = get_out_ip()
    # writelog("out_ip_list = {}".format(out_ip_list))

    for ip in out_ip_list:
        writelog("ping " + ip + " -c 1 -w 5 -I " + netcard)
        command = ["/usr/bin/ping", ip, "-c", "1", "-w", "5", "-I", netcard]
        try:
            result = subprocess.run(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
            if result.returncode == 0: #能ping通
                ret = 1
                break;
        except Exception as e:
            print("Error executing command:", e)

    writelog("网卡 {} 能否联通外网: {}".format(netcard, ret))
    return ret

def tag_netcard():
    """标记网卡"""
    netcard_dict = {} # <网卡名:标记> 字典
    dns_list = get_dns()
    netcard_list = get_devices()
    writelog("获取dns列表 dns_list = {}".format(dns_list))
    writelog("获取网卡列表 netcardList = {}".format(netcard_list))

    # 默认标记
    for netcard in netcard_list:
        netcard_dict[netcard] = "default"
    netcard_dict['lo'] = "local"

    # 标记vpn网卡
    net_object = SystemBus().get("org.freedesktop.NetworkManager", "/org/freedesktop/NetworkManager")
    devices = net_object.AllDevices
    for device in devices:
        device_object = SystemBus().get("org.freedesktop.NetworkManager",device)
        device_type = device_object.DeviceType
        device_interface = device_object.Interface
        # writelog("device = {}, device_interface = {}, device_type = {}".format(device, device_interface, device_type))
        if(device_type == 23):
            netcard_dict[device_interface] = "vpn"

    # 标记内网网卡
    dns_object = SystemBus().get("org.freedesktop.NetworkManager","/org/freedesktop/NetworkManager/DnsManager")
    dns_config = dns_object.Configuration
    writelog("dns_config = {}".format(dns_config))
    for i in dns_config:
        for j in i["nameservers"]:
            if(j in dns_list):# 如果系统网卡dns在配置的内网dns列表且能连通内网里，则该网卡标记为内网网卡
                if(check_in_net(i["interface"])):
                    netcard_dict[i["interface"]] = "in"
    
    for i in netcard_list:
        if netcard_dict[i]=="default":
            netcard_dict[i]="out"

    writelog("网卡标记 netcard_dict = {}".format(netcard_dict))
    return netcard_dict

def net_control(dbus_service):
    """断开外网网卡"""
    netcard_dict = tag_netcard()
    inflag = 0
    if "in" in netcard_dict.values() or "vpn" in netcard_dict.values():
        inflag = 1
        writelog("存在内网网卡")

    active_netcard = [] # 当前活跃的网卡
    net_object = SystemBus().get("org.freedesktop.NetworkManager", "/org/freedesktop/NetworkManager")
    active_cons = net_object.ActiveConnections
    writelog("所有活跃连接 active_cons = {}".format(active_cons))
    for i in active_cons:
        active_object = SystemBus().get("org.freedesktop.NetworkManager", i)
        active_devices = active_object.Devices
        writelog("当前活跃的设备 active_devices = {}".format(active_devices))
        for j in active_devices:
            state = SystemBus().get("org.freedesktop.NetworkManager", j).State
            active_object = SystemBus().get("org.freedesktop.NetworkManager",j)
            active_netcard = active_object.Interface
            writelog("当前活跃的网卡 active_netcard = {}, State = {}".format(active_netcard, state))
            if(state != 100):
                writelog("state不为100, 跳过")
                continue
            try:
                if inflag == 1 and netcard_dict[active_netcard] == "out" and os.path.isfile("/etc/net_isolation.conf") == True:
                    thread_net_disconnect = threading.Thread(target = disconnect_out_net,args=(dbus_service, active_object, active_netcard))
                    thread_net_disconnect.start()
            except Exception as e:
                writelog("error: {}".format(e))

def disconnect_out_net(dbus_service, active_object, active_netcard):
    if check_out_net(active_netcard):
        active_object.Disconnect()
        dbus_service.OnEmitSignal_netcard()
        passlog("根据我行信息安全管理要求，检测到您内外网同时连接，已断开您外网网卡 " + active_netcard + " 的网络连接。")

class Checkinoutnet(dbus.service.Object):
    def __init__(self):
        bus_name = dbus.service.BusName('com.kylin.checkinoutnet', bus = dbus.SystemBus())
        self.obj = dbus.service.Object.__init__(self, bus_name, '/com/kylin/checkinoutnet')

    @dbus.service.signal(dbus_interface='com.kylin.checkinoutnet')
    def OnEmitSignal_netcard(self):
        writelog("发送信号: OnEmitSignal_netcard")

net_status = get_status()
def timing_detection(dbus_service):
    """定时检测"""
    net_time = int(get_net_time())
    if(net_time == 0): 
        net_time = 600
    writelog("定时检测时间为 {}".format(net_time))
    if(net_status == 1):
        writelog("当前已开启管控,开始定时检测")
        thread_timing_detection = threading.Thread(target = net_control,args=(dbus_service,))
        thread_timing_detection.setDaemon(True)
        thread_timing_detection.start()
        time.sleep(net_time)
    
    t = Timer(net_time,timing_detection,(dbus_service,))
    t.start()

def signal_net_control_handler(*args, **kwargs):
    """接收管控信号"""
    writelog("收到管控信号")
    global net_status
    status = 0;
    if(args != None):
        status = args[0]
        writelog("管控状态 status = {}".format(status))
        if(status == 0):
            net_status = 0
            writelog("关闭管控，更改管控状态 net_status = {}".format(net_status))
        else:
            net_status = 1
            writelog("开启管控，更改管控状态 net_status = {}".format(net_status))
            thread_net_control = threading.Thread(target = net_control,args=(dbus_service,))
            thread_net_control.start()


def signal_state_change_handler(*args, **kwargs):
    """处理网络状态变化"""
    writelog("++++++++++++++++++++++++++++")
    state = args[0]
    writelog("state = {}".format(state))
    if state == 100 and net_status == 1:
        writelog("收到网卡已连接状态信号")
        writelog("开始重新检测......")
        thread_state_change_control = threading.Thread(target = net_control,args=(dbus_service,))
        thread_state_change_control.start()

def listening_net_state_change(device):
    """监听网络状态变化"""
    bus.add_signal_receiver(signal_state_change_handler,
                                        path = device,
                                        signal_name="StateChanged")

def signal_net_device_change_handler(*args, **kwargs):
    """处理网络设备变化"""
    writelog("args = {}".format(args))
    for i, arg in enumerate(args):
        device = args[i]
        writelog("add listen device = {}".format(device))
        thread_listen = threading.Thread(target = listening_net_state_change,args=(device,))
        thread_listen.setDaemon(True)
        thread_listen.start()

if __name__ == '__main__':
    DBusGMainLoop(set_as_default = True)
    dbus_service = Checkinoutnet()
    bus = dbus.SystemBus()

    # 监听管控信号
    bus.add_signal_receiver(signal_net_control_handler,
                            dbus_interface = "com.kylin.netIsolationctl",
                            signal_name = "NetControlSignal")

    # 监听网络设备变化
    bus.add_signal_receiver(signal_net_device_change_handler,
                                    dbus_interface = "org.freedesktop.NetworkManager",
                                    signal_name = "DeviceAdded")

    # 监听网络状态变化
    net_object = SystemBus().get("org.freedesktop.NetworkManager", "/org/freedesktop/NetworkManager")
    devices = net_object.AllDevices
    for device in devices:
        device_object = SystemBus().get("org.freedesktop.NetworkManager", device)
        netcard = device_object.Interface
        writelog("init listen device = {}, netcard = {}".format(device, netcard))
        thread_listen = threading.Thread(target = listening_net_state_change,args=(device,))
        thread_listen.setDaemon(True)
        thread_listen.start()

    #定时检测
    thread_timing = threading.Thread(target = timing_detection,args=(dbus_service,))
    thread_timing.setDaemon(True)
    thread_timing.start()

    mainloop = GLib.MainLoop()
    mainloop.run()
    
